from functools import cached_property

import dagster as dg
from dagster._annotations import preview, public
from pydantic import Field

from dagster_aws.components.credentials import Boto3CredentialsComponent
from dagster_aws.ssm.resources import ParameterStoreResource, ParameterStoreTag, SSMResource


@public
@preview
class SSMResourceComponent(dg.Component, dg.Resolvable, dg.Model):
    """A component that provides a SSMResource for interacting with AWS Systems Manager."""

    credentials: Boto3CredentialsComponent = Field(
        description="AWS credentials - inline configuration."
    )
    resource_key: str | None = Field(
        default=None, description="The key under which the resource will be bound in definitions."
    )

    @cached_property
    def resource(self) -> SSMResource:
        """Resolves credentials and returns a configured SSM resource."""
        return SSMResource(**self.credentials.render_as_dict())

    def build_defs(self, context: dg.ComponentLoadContext) -> dg.Definitions:
        if self.resource_key is None:
            return dg.Definitions()
        return dg.Definitions(resources={self.resource_key: self.resource})


@public
@preview
class ParameterStoreResourceComponent(dg.Component, dg.Resolvable, dg.Model):
    """A component that provides a ParameterStoreResource for fetching parameters from AWS SSM Parameter Store."""

    credentials: Boto3CredentialsComponent = Field(
        description="AWS credentials - inline configuration."
    )

    parameters: list[str] = Field(
        default=[],
        description="AWS SSM Parameter store parameters with this tag will be fetched and made available.",
    )

    parameter_tags: list[ParameterStoreTag] = Field(
        default=[],
        description="AWS SSM Parameter store parameters with this tag will be fetched and made available.",
    )

    parameter_paths: list[str] = Field(
        default=[], description="List of path prefixes to pull parameters from."
    )
    with_decryption: bool = Field(
        default=False, description="Whether to decrypt secure string parameters."
    )

    resource_key: str | None = Field(
        default=None,
        description="The key under which the ParameterStore resource will be bound to the definitions.",
    )

    @cached_property
    def resource(self) -> ParameterStoreResource:
        return ParameterStoreResource(
            **self.credentials.render_as_dict(),
            parameters=self.parameters,
            parameter_paths=self.parameter_paths,
            parameter_tags=self.parameter_tags,
            with_decryption=self.with_decryption,
        )

    def build_defs(self, context: dg.ComponentLoadContext) -> dg.Definitions:
        if self.resource_key is None:
            return dg.Definitions()
        return dg.Definitions(resources={self.resource_key: self.resource})



from collections.abc import Generator
from contextlib import contextmanager
from typing import TYPE_CHECKING, cast

from dagster import (
    Field as LegacyDagsterField,
    resource,
)
from dagster._annotations import beta
from dagster._core.definitions.resource_definition import dagster_maintained_resource
from dagster._core.test_utils import environ
from dagster._utils.merger import merge_dicts
from pydantic import Field

from dagster_aws.secretsmanager.secrets import (
    construct_secretsmanager_client,
    get_secrets_from_arns,
    get_tagged_secrets,
)
from dagster_aws.utils import ResourceWithBoto3Configuration

if TYPE_CHECKING:
    import botocore
    from dagster._config.field_utils import Shape


@beta
class SecretsManagerResource(ResourceWithBoto3Configuration):
    """Resource that gives access to AWS SecretsManager.

    The underlying SecretsManager session is created by calling
    :py:func:`boto3.session.Session(profile_name) <boto3:boto3.session>`.
    The returned resource object is a SecretsManager client, an instance of `botocore.client.SecretsManager`.

    Example:
        .. code-block:: python

            from dagster import build_op_context, job, op
            from dagster_aws.secretsmanager import SecretsManagerResource

            @op
            def example_secretsmanager_op(secretsmanager: SecretsManagerResource):
                return secretsmanager.get_client().get_secret_value(
                    SecretId='arn:aws:secretsmanager:region:aws_account_id:secret:appauthexample-AbCdEf'
                )

            @job
            def example_job():
                example_secretsmanager_op()

            Definitions(
                jobs=[example_job],
                resources={
                    'secretsmanager': SecretsManagerResource(
                        region_name='us-west-1'
                    )
                }
            )
    """

    @classmethod
    def _is_dagster_maintained(cls) -> bool:
        return True

    def get_client(self) -> "botocore.client.SecretsManager":  # pyright: ignore (reportAttributeAccessIssue)
        return construct_secretsmanager_client(
            max_attempts=self.max_attempts,
            region_name=self.region_name,
            profile_name=self.profile_name,
            endpoint_url=self.endpoint_url,
            use_ssl=self.use_ssl,
            verify=self.verify,
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key,
            aws_session_token=self.aws_session_token,
        )


@beta
@dagster_maintained_resource
@resource(SecretsManagerResource.to_config_schema())
def secretsmanager_resource(context) -> "botocore.client.SecretsManager":  # pyright: ignore (reportAttributeAccessIssue)
    """Resource that gives access to AWS SecretsManager.

    The underlying SecretsManager session is created by calling
    :py:func:`boto3.session.Session(profile_name) <boto3:boto3.session>`.
    The returned resource object is a SecretsManager client, an instance of `botocore.client.SecretsManager`.

    Example:
        .. code-block:: python

            from dagster import build_op_context, job, op
            from dagster_aws.secretsmanager import secretsmanager_resource

            @op(required_resource_keys={'secretsmanager'})
            def example_secretsmanager_op(context):
                return context.resources.secretsmanager.get_secret_value(
                    SecretId='arn:aws:secretsmanager:region:aws_account_id:secret:appauthexample-AbCdEf'
                )

            @job(resource_defs={'secretsmanager': secretsmanager_resource})
            def example_job():
                example_secretsmanager_op()

            example_job.execute_in_process(
                run_config={
                    'resources': {
                        'secretsmanager': {
                            'config': {
                                'region_name': 'us-west-1',
                            }
                        }
                    }
                }
            )

    You may configure this resource as follows:

    .. code-block:: YAML

        resources:
          secretsmanager:
            config:
              region_name: "us-west-1"
              # Optional[str]: Specifies a custom region for the SecretsManager session. Default is chosen
              # through the ordinary boto credential chain.
              profile_name: "dev"
              # Optional[str]: Specifies a custom profile for SecretsManager session. Default is default
              # profile as specified in ~/.aws/credentials file

    """
    return SecretsManagerResource.from_resource_context(context).get_client()


@beta
class SecretsManagerSecretsResource(ResourceWithBoto3Configuration):
    """Resource that provides a dict which maps selected SecretsManager secrets to
    their string values. Also optionally sets chosen secrets as environment variables.

    Example:
        .. code-block:: python

            import os
            from dagster import build_op_context, job, op, ResourceParam
            from dagster_aws.secretsmanager import SecretsManagerSecretsResource

            @op
            def example_secretsmanager_secrets_op(secrets: SecretsManagerSecretsResource):
                return secrets.fetch_secrets().get("my-secret-name")

            @op
            def example_secretsmanager_secrets_op_2(secrets: SecretsManagerSecretsResource):
                with secrets.secrets_in_environment():
                    return os.getenv("my-other-secret-name")

            @job
            def example_job():
                example_secretsmanager_secrets_op()
                example_secretsmanager_secrets_op_2()

            Definitions(
                jobs=[example_job],
                resources={
                    'secrets': SecretsManagerSecretsResource(
                        region_name='us-west-1',
                        secrets_tag="dagster",
                        add_to_environment=True,
                    )
                }
            )

    Note that your ops must also declare that they require this resource with or it will not be initialized
    for the execution of their compute functions.
    """

    secrets: list[str] = Field(
        default=[], description="An array of AWS Secrets Manager secrets arns to fetch."
    )
    secrets_tag: str | None = Field(
        default=None,
        description="AWS Secrets Manager secrets with this tag will be fetched and made available.",
    )

    @classmethod
    def _is_dagster_maintained(cls) -> bool:
        return True

    @contextmanager
    def secrets_in_environment(
        self,
        secrets: list[str] | None = None,
        secrets_tag: str | None = None,
    ) -> Generator[dict[str, str], None, None]:
        """Yields a dict which maps selected SecretsManager secrets to their string values. Also
        sets chosen secrets as environment variables.

        Args:
            secrets (Optional[List[str]]): An array of AWS Secrets Manager secrets arns to fetch.
                Note that this will override the secrets specified in the resource config.
            secrets_tag (Optional[str]): AWS Secrets Manager secrets with this tag will be fetched
                and made available. Note that this will override the secrets_tag specified in the
                resource config.
        """
        secrets_manager = construct_secretsmanager_client(
            max_attempts=self.max_attempts,
            region_name=self.region_name,
            profile_name=self.profile_name,
        )

        secrets_tag_to_fetch = secrets_tag if secrets_tag is not None else self.secrets_tag
        secrets_to_fetch = secrets if secrets is not None else self.secrets

        secret_arns = merge_dicts(
            (
                get_tagged_secrets(secrets_manager, [secrets_tag_to_fetch])
                if secrets_tag_to_fetch
                else {}
            ),
            get_secrets_from_arns(secrets_manager, secrets_to_fetch),
        )

        secrets_map = {
            name: secrets_manager.get_secret_value(SecretId=arn).get("SecretString")
            for name, arn in secret_arns.items()
        }
        with environ(secrets_map):
            yield secrets_map

    def fetch_secrets(
        self,
        secrets: list[str] | None = None,
        secrets_tag: str | None = None,
    ) -> dict[str, str]:
        """Fetches secrets from AWS Secrets Manager and returns them as a dict.

        Args:
            secrets (Optional[List[str]]): An array of AWS Secrets Manager secrets arns to fetch.
                Note that this will override the secrets specified in the resource config.
            secrets_tag (Optional[str]): AWS Secrets Manager secrets with this tag will be fetched
                and made available. Note that this will override the secrets_tag specified in the
                resource config.
        """
        with self.secrets_in_environment(secrets=secrets, secrets_tag=secrets_tag) as secret_values:
            return secret_values


LEGACY_SECRETSMANAGER_SECRETS_SCHEMA = {
    **cast("Shape", SecretsManagerSecretsResource.to_config_schema().as_field().config_type).fields,
    "add_to_environment": LegacyDagsterField(
        bool,
        default_value=False,
        description="Whether to add the secrets to the environment. Defaults to False.",
    ),
}


@beta
@dagster_maintained_resource
@resource(config_schema=LEGACY_SECRETSMANAGER_SECRETS_SCHEMA)
@contextmanager
def secretsmanager_secrets_resource(context):
    """Resource that provides a dict which maps selected SecretsManager secrets to
    their string values. Also optionally sets chosen secrets as environment variables.

    Example:
        .. code-block:: python

            import os
            from dagster import build_op_context, job, op
            from dagster_aws.secretsmanager import secretsmanager_secrets_resource

            @op(required_resource_keys={'secrets'})
            def example_secretsmanager_secrets_op(context):
                return context.resources.secrets.get("my-secret-name")

            @op(required_resource_keys={'secrets'})
            def example_secretsmanager_secrets_op_2(context):
                return os.getenv("my-other-secret-name")

            @job(resource_defs={'secrets': secretsmanager_secrets_resource})
            def example_job():
                example_secretsmanager_secrets_op()
                example_secretsmanager_secrets_op_2()

            example_job.execute_in_process(
                run_config={
                    'resources': {
                        'secrets': {
                            'config': {
                                'region_name': 'us-west-1',
                                'secrets_tag': 'dagster',
                                'add_to_environment': True,
                            }
                        }
                    }
                }
            )

    Note that your ops must also declare that they require this resource with
    `required_resource_keys`, or it will not be initialized for the execution of their compute
    functions.

    You may configure this resource as follows:

    .. code-block:: YAML

        resources:
          secretsmanager:
            config:
              region_name: "us-west-1"
              # Optional[str]: Specifies a custom region for the SecretsManager session. Default is chosen
              # through the ordinary boto credential chain.
              profile_name: "dev"
              # Optional[str]: Specifies a custom profile for SecretsManager session. Default is default
              # profile as specified in ~/.aws/credentials file
              secrets: ["arn:aws:secretsmanager:region:aws_account_id:secret:appauthexample-AbCdEf"]
              # Optional[List[str]]: Specifies a list of secret ARNs to pull from SecretsManager.
              secrets_tag: "dagster"
              # Optional[str]: Specifies a tag, all secrets which have the tag set will be pulled
              # from SecretsManager.
              add_to_environment: true
              # Optional[bool]: Whether to set the selected secrets as environment variables. Defaults
              # to false.

    """
    add_to_environment = context.resource_config.get("add_to_environment", False)
    if add_to_environment:
        with SecretsManagerSecretsResource.from_resource_context(
            context
        ).secrets_in_environment() as secrets:
            yield secrets
    else:
        yield SecretsManagerSecretsResource.from_resource_context(context).fetch_secrets()



from functools import cached_property

import dagster as dg
from dagster._annotations import preview, public
from pydantic import BaseModel, Field

from dagster_aws.redshift.resources import RedshiftClientResource


class CredentialsRenderMixin:
    """A mixin to provide shared dictionary rendering logic for AWS credentials."""

    def render_as_dict(self) -> dict:
        """Returns the credentials as a dictionary, excluding None values."""
        assert isinstance(self, BaseModel)
        return self.model_dump(exclude_none=True)


@public
@preview
class RedshiftCredentialsComponent(dg.Component, dg.Resolvable, dg.Model, CredentialsRenderMixin):
    """Credentials and connection configuration for Redshift."""

    host: str | None = Field(default=None, description="Redshift host")
    port: int = Field(default=5439, description="Redshift port")
    user: str | None = Field(default=None, description="Username for Redshift connection")
    password: str | None = Field(default=None, description="Password for Redshift connection")
    database: str | None = Field(
        default=None,
        description=(
            "Name of the default database to use. After login, you can use USE DATABASE to change"
            " the database."
        ),
    )
    autocommit: bool | None = Field(default=None, description="Whether to autocommit queries")
    connect_timeout: int = Field(
        default=5, description="Timeout for connection to Redshift cluster. Defaults to 5 seconds."
    )
    sslmode: str | None = Field(
        default="require",
        description=(
            "SSL mode to use. See the Redshift documentation for reference:"
            " https://docs.aws.amazon.com/redshift/latest/mgmt/connecting-ssl-support.html"
        ),
    )

    def build_defs(self, context: dg.ComponentLoadContext) -> dg.Definitions:
        return dg.Definitions()


@public
@preview
class RedshiftClientResourceComponent(dg.Component, dg.Resolvable, dg.Model):
    """A component that provides a RedshiftClientResource for connecting to and querying Amazon Redshift."""

    credentials: RedshiftCredentialsComponent = Field(
        description="Redshift credentials - inline configuration."
    )
    resource_key: str | None = Field(
        default=None,
        description="The key under which the Redshift resource will be bound to the definitions.",
    )

    @cached_property
    def resource(self) -> RedshiftClientResource:
        """Resolves credentials and returns a configured Redshift resource."""
        return RedshiftClientResource(**self.credentials.render_as_dict())

    def build_defs(self, context: dg.ComponentLoadContext) -> dg.Definitions:
        if self.resource_key is None:
            return dg.Definitions()
        return dg.Definitions(resources={self.resource_key: self.resource})



import dagster as dg
from dagster._annotations import preview, public
from pydantic import BaseModel, Field


class CredentialsRenderMixin:
    """A mixin to provide shared dictionary rendering logic for AWS credentials."""

    def render_as_dict(self) -> dict:
        """Returns the credentials as a dictionary, excluding None values."""
        assert isinstance(self, BaseModel)
        return self.model_dump(exclude_none=True)


@public
@preview
class Boto3CredentialsComponent(dg.Component, dg.Resolvable, dg.Model, CredentialsRenderMixin):
    """Configuration for standard AWS SDK (Boto3) credentials and session settings."""

    region_name: str | None = Field(
        default=None, description="Specifies a custom region for the Boto3 session"
    )
    endpoint_url: str | None = Field(
        default=None, description="Specifies a custom endpoint for the Boto3 session."
    )
    aws_access_key_id: str | None = Field(
        default=None, description="AWS access key ID to use when creating the boto3 session."
    )
    aws_secret_access_key: str | None = Field(
        default=None, description="AWS secret access key to use when creating the boto3 session."
    )
    aws_session_token: str | None = Field(
        default=None, description="AWS session token to use when creating the boto3 session."
    )
    profile_name: str | None = Field(
        default=None, description="Specifies a profile to connect that session"
    )
    use_ssl: bool = Field(
        default=True, description="Whether or not to use SSL. By default, SSL is used."
    )
    verify: bool | None = Field(
        default=True,
        description="Whether or not to verify SSL certificates. By default SSL certificates are verified.",
    )
    max_attempts: int = Field(
        default=5,
        description=(
            "This provides Boto3's retry handler with a value of maximum retry attempts, where the"
            " initial call counts toward the max_attempts value that you provide"
        ),
    )
    retry_mode: str | None = Field(
        default=None, description="Specifies the retry mode to use for the Boto3 session."
    )

    def build_defs(self, context: dg.ComponentLoadContext) -> dg.Definitions:
        return dg.Definitions()


@public
@preview
class S3CredentialsComponent(Boto3CredentialsComponent):
    """Configuration for S3-specific credentials, including unsigned session support."""

    use_unsigned_session: bool | None = Field(
        default=None, description="Specifies whether to use an unsigned S3 session."
    )


@public
@preview
class AthenaCredentialsComponent(Boto3CredentialsComponent):
    """Configuration for Athena-specific authentication and query settings."""

    workgroup: str | None = Field(
        default="primary",
        description=(
            "The Athena WorkGroup to use."
            " https://docs.aws.amazon.com/athena/latest/ug/manage-queries-control-costs-with-workgroups.html"
        ),
    )
    polling_interval: int = Field(
        default=5,
        description=(
            "Time in seconds between checks to see if a query execution is finished. 5 seconds"
            " by default. Must be non-negative."
        ),
    )
    max_polls: int = Field(
        default=120,
        description=(
            "Number of times to poll before timing out. 120 attempts by default. When coupled"
            " with the default polling_interval, queries will timeout after 10 minutes (120 * 5"
            " seconds). Must be greater than 0."
        ),
    )



from dagster_aws.ssm.resources import (
    ParameterStoreResource as ParameterStoreResource,
    ParameterStoreTag as ParameterStoreTag,
    SSMResource as SSMResource,
    parameter_store_resource as parameter_store_resource,
    ssm_resource as ssm_resource,
)
